"""
Classes to implement qasm2 translation to python data structures.

Notes
-----
This is a modified version of :mod: `~nuqasm2.qasmast`. The modifications are
mainly to convert from the command line interface used by
:mod:`~nuqasm2.qasmast` to the typical python interface.

Some of the code is also adapted from the examples module in the pyparsing
package
"""

# =============================================================================
# Derived from qasmast.py.
# Classes to implement qasm2 translation to python data structures
# Working from _Open Quantum Assembly Language_
# https://arxiv.org/pdf/1707.03429.pdf
# jack j. woehr jwoehr@softwoehr.com po box 51 golden co 80402-0051
# Copyright 2019 Jack Woehr jwoehr@softwoehr.com PO Box 51, Golden, CO 80402-0051.
# Apache-2.0 license -- See LICENSE which you should have received with this code.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
#
# MODIFIED ON 10/10/2023 BY KENNY CAMPBELL elkmc@leeds.ac.uk. I have added the
# qasm2ast function to wrap some of the functionality in a way convenient for my
# use case, which is use of the abstract sytax tree (ast) within python rather
# than accessing it via the command line. The qasm2ast function is modified from
# the do_it function in the nuqasm2/nuqasm2.py script, which has the same
# copyright info and licensing as above. This useage is in accordance
# with the apache 2.0 license available at
# licenses_of_code_parts_of_package_are_derived_from/LICENSE4qasm2ast_base_code
# or https://www.apache.org/licenses/LICENSE-2.0
# =============================================================================

from enum import Enum
import re
import datetime
import os
from functools import wraps

import pyparsing as pp


class QTRegEx:
    """Compiled regular expressions for parsing."""

    COMMENT = re.compile(r"^\s*//")
    INCLUDE = re.compile(r"^\s*include\s+\"\S+\"\s*;")
    EOL_COMMENT = re.compile(r";.*(//.*)")

    CTL_2 = re.compile(r"(\w+)\((\w+)(\W+)(\w+)\)\s+(\w+)\s+(\w+\[\w+\]);")

    QREG = re.compile(r"^\s*qreg\s+\S*\[\d+\]\s*;")
    CREG = re.compile(r"^\s*creg\s+\S*\[\d+\]\s*;")
    MEASURE = re.compile(r"^\s*measure\s+\S+\s*\-\>\s*\S+\s*;")
    BARRIER = re.compile(r"^\s*barrier\s+.*;")
    GATE = re.compile(r"^\s*gate\s+.*")
    # OP = re.compile(r"^\s*\S+\s+\S+\[\d+\]\s*;")
    OP = re.compile(r"^\s*\S+\s+\S+\s*;")

    INCLUDE_TARGET = re.compile(r".*\"(\S+)\"\s*;")
    REG_DECL = re.compile(r".*(\S+)\[(\d+)\].*")
    MEASURE_DECL = re.compile(r"^\s*measure\s+(\S+)\s*\-\>\s*(\S+)\s*;")
    BARRIER_DECL = re.compile(r"\S+\[\d+\]")
    BARRIER_DECL_1 = re.compile(r"\S+\s*[\,;]")
    OP_AND_ARGS = re.compile(r"^\s*(\S+)\s+.*")
    OP_PARAM_LIST = re.compile(r"(\S+)\((.*)\)")
    # OP_REG_LIST = re.compile(r"\S+\[\d+\]")
    OP_REG_LIST = re.compile(r"\s+\S+;")

    START_CURLY = re.compile(r".*\{.*")
    END_CURLY = re.compile(r".*\}.*")
    GATE_DECL = re.compile(r"gate\s+(\S+)\s+(\S*).*")
    GATE_OPS = re.compile(r"gate\s+.*\{(.*)\}")
    GATE_OPS_LIST = re.compile(r"\S+\s+\S+;")
    GATE_OP = re.compile(r"(\w+).*")
    GATE_OP_PARAMS = re.compile(r"\S+\((.*)\).*")
    GATE_OP_REGS = re.compile(r".*\s+(\S+);")  # any characters followed by at least
    # one whitespace followed by at
    # least one non-whitespace character


# useful parsing terms
class QasmParsingElement:
    """ParsingElement defined using the pyparsing package with some additional
    more specific methods. I have added this because the RegEx expressions
    used in the original nuqasm2 code are leading to bugs. For example, only
    the first character of qreg names are being stored under the qreg_name key
    of the ASTType.QREG entries in the AST c_sect."""

    def __init__(self):
        pass

    lpar = pp.Literal("(").suppress()
    rpar = pp.Literal(")").suppress()
    l_sqr_brace = pp.Literal("[").suppress()
    r_sqr_brace = pp.Literal("]").suppress()
    idQasm = pp.Regex(r"[a-z][A-Za-z0-9]*")  # id from https://arxiv.org/abs/1707.03429
    real = pp.Regex(r"([0-9]+\.[0-9]*|[0-9]*\.[0-9]+)([eE][-+]?[0-9]+)?")
    nninteger = pp.Regex(r"[1-9]+[0-9]*|0")
    number = real | nninteger | pp.Keyword("pi")
    arith_op = pp.one_of(["+", "-", "*", "/", "^"])
    unaryop = pp.one_of(["sin", "cos", "tan", "exp", "ln", "sqrt"])
    reg_index_slice = l_sqr_brace + nninteger + r_sqr_brace
    decl = (
        pp.Keyword("qreg") + idQasm + reg_index_slice
        | pp.Keyword("creg") + idQasm + reg_index_slice
    )


# =============================================================================
# # =============================================================================
# #     exp_num_or_word = real | nninteger | pi | idQasm
# # =============================================================================
# # =============================================================================
# #     params =  pp.original_text_for(pp.nested_expr) #anything (including more
# # =============================================================================
#                                                    #nested brackets) inside
#                                                    #parentheses
# # =============================================================================
# #     atom = number | pp.Keyword('pi')
# #     exp_ops = [('-', 1, pp.OpAssoc.LEFT), ('^', 2, pp.OpAssoc.RIGHT),
# #                  (pp.one_of(['*', '/']), 2, pp.OpAssoc.LEFT),
# #                  (pp.one_of(['+', '-']), 2, pp.OpAssoc.LEFT),
# #                  (unaryop, 1, pp.OpAssoc.LEFT)]
# #     exp_qasm = pp.infix_notation(atom, exp_ops, '(', ')')
# # =============================================================================
#     # use CaselessKeyword for e and pi, to avoid accidentally matching
#     # functions that start with 'e' or 'pi' (such as 'exp'); Keyword
#     # and CaselessKeyword only match whole words
# =============================================================================


class ASTType(Enum):
    """
    The line types that can be handled by the parser.

    Parameters
    ----------
    source : str
        A line from the  source code (.qasm file) to be parsed.

    Returns
    -------
    :class:`~enum.Enum`
        Enumerator indicating line type.
    """

    UNKNOWN = 0
    COMMENT = 100
    QREG = 20
    CREG = 30
    MEASURE = 40
    BARRIER = 50
    GATE = 60
    OP = 70
    CTL = 80
    CTL_2 = 82
    BLANK = 1000
    DECLARATION_QASM_2_0 = 2000
    INCLUDE = 3000

    @classmethod
    def astType(cls, source):
        """
        Return enum indicating line type
        source is source code
        """
        if source == "":
            return cls.BLANK
        if source == "OPENQASM 2.0;":
            return cls.DECLARATION_QASM_2_0
        x = QTRegEx.COMMENT.search(source)
        if x:
            return cls.COMMENT
        x = QTRegEx.INCLUDE.search(source)
        if x:
            return cls.INCLUDE
        x = QTRegEx.CTL_2.search(source)
        if x:
            if x.group(1) == "if":
                return cls.CTL_2
        x = QTRegEx.QREG.search(source)
        if x:
            return cls.QREG
        x = QTRegEx.CREG.search(source)
        if x:
            return cls.CREG
        x = QTRegEx.MEASURE.search(source)
        if x:
            return cls.MEASURE
        x = QTRegEx.BARRIER.search(source)
        if x:
            return cls.BARRIER
        x = QTRegEx.GATE.search(source)
        if x:
            return cls.GATE
        x = QTRegEx.OP.search(source)
        if x:
            return cls.OP
        return cls.UNKNOWN

    @classmethod
    def ast_eol_comment(cls, source):
        """
        Handles comments in source code

        Parameters
        ----------
        source : str
            A line from the  source code (.qasm file) to be parsed.

        Returns
        -------
        x : str
            The commented line with leading and trailing spaces removed.
        """
        x = QTRegEx.EOL_COMMENT.search(source)
        if x:
            x = x.group(1).strip()
        return x


class ASTElement:
    """
    A base class for AST elements.

    Parameters
    ----------
    filenum : int
        Index of filepath in :class:`T_Sect`
        filepaths vector.
    linenum : int
        The linenumber of the source code line being parsed.
    ast_type : element of :class:`ASTType`
    source : str
        A line from the  source code (.qasm file) to be parsed.
    save_element_source : bool
        Whether to save the relevant unparsed source code line or not. Default
        is False.
    eol_comment : str or None
        The end of line comment (if there is one).
    """

    def _eol_comment(f):
        """Add another element to the AST if an end-of-line comment present"""

        @wraps(f)
        def wrapped(inst):
            ast = f(inst)
            if inst.eol_comment:
                ast["eol_comment"] = inst.eol_comment
            return ast

        return wrapped

    def __init__(
        self,
        filenum,
        linenum,
        ast_type,
        source,
        save_element_source=False,
        eol_comment=None,
    ):
        """Instance from qasm source code and parse into key:value pairs"""
        self.filenum = filenum
        self.linenum = linenum
        self.ast_type = ast_type
        self.source = source
        self.save_element_source = save_element_source
        self.eol_comment = eol_comment

    @_eol_comment
    def out(self):
        """Returns self as a dict structure"""
        return {
            "filenum": self.filenum,
            "linenum": self.linenum,
            "type": self.ast_type,
            "source": self.source if self.save_element_source else None,
        }

    @staticmethod
    def proc_reg_list(txt):
        """Internal parsing routine for reg list of ops"""
        x = QTRegEx.OP_REG_LIST.findall(txt)
        y = x[0].strip(";")
        y = y.strip()
        return y.split(",")


class ASTElementUnknown(ASTElement):
    """
    ASTElementUnknown
    An unknown ASTElement, unimplemented or error
    Knows linenum, ast_type, source
    """

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementUnknown, self).__init__(
            filenum, linenum, ASTType.UNKNOWN, source, save_element_source, eol_comment
        )


class ASTElementComment(ASTElement):
    """
    ASTElementComment
    A programmer comment element
    Knows linenum, ast_type, source
    """

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementComment, self).__init__(
            filenum, linenum, ASTType.COMMENT, source, save_element_source, eol_comment
        )


class ASTElementDeclarationQasm2_0(ASTElement):
    """
    ASTElementDeclarationQasm2_0
    Obligatory declaration
    Knows linenum, ast_type, source
    """

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementDeclarationQasm2_0, self).__init__(
            filenum,
            linenum,
            ASTType.DECLARATION_QASM_2_0,
            source,
            save_element_source,
            eol_comment,
        )


class ASTElementInclude(ASTElement):
    """
    For handling openQASM2 include statements.
    Knows linenum, ast_type, source, include
    """

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementInclude, self).__init__(
            filenum, linenum, ASTType.INCLUDE, source, save_element_source, eol_comment
        )
        x = QTRegEx.INCLUDE_TARGET.search(source)
        self.include = x.group(1)

    @ASTElement._eol_comment
    def out(self):
        return {
            "filenum": self.filenum,
            "linenum": self.linenum,
            "type": self.ast_type,
            "source": self.source if self.save_element_source else None,
            "include": self.include,
        }


class ASTElementQReg(ASTElement):
    """
    For handling qreg declarations.
    Knows linenum, ast_type, source, qreg_name, qreg_num
    """

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementQReg, self).__init__(
            filenum, linenum, ASTType.QREG, source, save_element_source, eol_comment
        )
        parsed_line = QasmParsingElement.decl.parse_string(self.source)
        self.qreg_name = parsed_line[1]
        self.qreg_num = parsed_line[2]
        # commented out block below is the original nuqasm2 code. It fails to
        # correctly parse multi-character register names and instead only
        # retrieves the last character of the name

    # =============================================================================
    #         x = QTRegEx.REG_DECL.match(self.source)
    #         self.qreg_name = x.group(1)
    #         self.qreg_num = x.group(2)
    # =============================================================================

    @ASTElement._eol_comment
    def out(self):
        return {
            "filenum": self.filenum,
            "linenum": self.linenum,
            "type": self.ast_type,
            "source": self.source if self.save_element_source else None,
            "qreg_name": self.qreg_name,
            "qreg_num": self.qreg_num,
        }


class ASTElementCReg(ASTElement):
    """
    For handling creg declarations.
    Knows linenum, ast_type, source, creg_name, creg_num
    """

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementCReg, self).__init__(
            filenum, linenum, ASTType.CREG, source, save_element_source, eol_comment
        )
        parsed_line = QasmParsingElement.decl.parse_string(self.source)
        self.creg_name = parsed_line[1]
        self.creg_num = parsed_line[2]
        # commented out block below is the original nuqasm2 code. It fails to
        # correctly parse multi-character register names and instead only
        # retrieves the last character of the name

    # =============================================================================
    #         x = QTRegEx.REG_DECL.match(self.source)
    #         self.creg_name = x.group(1)
    #         self.creg_num = x.group(2)
    # =============================================================================

    @ASTElement._eol_comment
    def out(self):
        return {
            "filenum": self.filenum,
            "linenum": self.linenum,
            "type": self.ast_type,
            "source": self.source if self.save_element_source else None,
            "creg_name": self.creg_name,
            "creg_num": self.creg_num,
        }


class ASTElementMeasure(ASTElement):
    """
    For handing measure commands.
    Knows linenum, ast_type, source, source_reg, target_reg
    """

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementMeasure, self).__init__(
            filenum, linenum, ASTType.MEASURE, source, save_element_source, eol_comment
        )
        x = QTRegEx.MEASURE_DECL.match(self.source)
        self.source_reg = x.group(1)
        self.target_reg = x.group(2)

    @ASTElement._eol_comment
    def out(self):
        return {
            "filenum": self.filenum,
            "linenum": self.linenum,
            "type": self.ast_type,
            "source": self.source if self.save_element_source else None,
            "source_reg": self.source_reg,
            "target_reg": self.target_reg,
        }


class ASTElementBarrier(ASTElement):
    """
    For handling barrier commands.
    Knows linenum, ast_type, source, reg_list
    """

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementBarrier, self).__init__(
            filenum, linenum, ASTType.BARRIER, source, save_element_source, eol_comment
        )
        x = QTRegEx.BARRIER_DECL.findall(self.source)
        if not x:  # e.g., qiskit-terra/examples/qasm/entangled_registers.qasm
            x = QTRegEx.BARRIER_DECL_1.findall(self.source)
            x[0] = x[0].rstrip(";")
        self.reg_list = x[0].split(",")

    @ASTElement._eol_comment
    def out(self):
        return {
            "filenum": self.filenum,
            "linenum": self.linenum,
            "type": self.ast_type,
            "source": self.source if self.save_element_source else None,
            "reg_list": self.reg_list,
        }


class ASTElementOp(ASTElement):
    """
    For handling openQASM2 gate operations and subgate operations (AKA operators).
    Knows linenum, ast_type, source, op, param_list, reg_list
    """

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementOp, self).__init__(
            filenum, linenum, ASTType.OP, source, save_element_source, eol_comment
        )
        x = QTRegEx.OP_AND_ARGS.match(self.source)
        op_and_args = x.group(1)
        x = QTRegEx.OP_PARAM_LIST.match(op_and_args)

        self.param_list = None
        if x:
            self.op = x.group(1)
            self.param_list = x.group(2).split(",")
        else:
            self.op = op_and_args

        self.reg_list = self.proc_reg_list(self.source)

    @ASTElement._eol_comment
    def out(self):
        return {
            "filenum": self.filenum,
            "linenum": self.linenum,
            "type": self.ast_type,
            "source": self.source if self.save_element_source else None,
            "op": self.op,
            "param_list": self.param_list,
            "reg_list": self.reg_list,
        }


class ASTElementCtl2(ASTElement):
    """
    Control-flow with binary operator.
    Knows linenum, ast_type, source, ctl, expression_op,
    expression_param_list, op, param_list, reg_list
    """

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementCtl2, self).__init__(
            filenum, linenum, ASTType.CTL_2, source, save_element_source, eol_comment
        )
        x = QTRegEx.CTL_2.match(self.source)
        self.ctl = x.group(1)
        self.expression_op = x.group(3)
        self.expression_param_list = [x.group(2), x.group(4)]

        op_and_args = x.group(5)
        x = QTRegEx.OP_PARAM_LIST.match(op_and_args)

        self.param_list = None
        if x:
            self.op = x.group(1)
            self.param_list = x.group(2).split(",")
        else:
            self.op = op_and_args

        self.reg_list = self.proc_reg_list(self.source)

    @ASTElement._eol_comment
    def out(self):
        return {
            "filenum": self.filenum,
            "linenum": self.linenum,
            "type": self.ast_type,
            "source": self.source if self.save_element_source else None,
            "ctl": self.ctl,
            "expression_op": self.expression_op,
            "expression_param_list": self.expression_param_list,
            "op": self.op,
            "param_list": self.param_list,
            "reg_list": self.reg_list,
        }


class ASTElementGateDefinitionPlaceholder(ASTElement):
    """So something will show up in the c_sect when a gate definition starts"""

    def __init__(
        self, filenum, linenum, source, save_element_source=False, eol_comment=None
    ):
        super(ASTElementGateDefinitionPlaceholder, self).__init__(
            filenum, linenum, ASTType.GATE, None, None
        )

    @ASTElement._eol_comment
    def out(self):
        """Returns self as a dict structure"""
        return {"filenum": self.filenum, "linenum": self.linenum, "type": self.ast_type}


class Gate_Operation:
    """
    One operation in a gate operation list.

    Parameters
    ----------
    op : str
        The gate name.
    op_param_list : list of str
        List of parameter values for the gate.
    op_reg_list : list of str
        The names of the quantum registers to act on.

    Attributes
    ----------
    operation : dict
        Dict of the input parameters.

    Notes
    -----
    I think that there is an issue with the output of op_reg_list inherrited
    from the original :mod: `~nuqasm2.qasmast`
    code. The output is the name of the declared qreg for native openQASM 2
    gates but is the name of the qubit arguments given for gates defined in
    standard header.
    """

    def __init__(self, op, op_param_list, op_reg_list):
        """
        Init operation step.

        Instance structures filled in by :class:`QasmTranslator`
        op
             operator
        op_param_list : list
            parameter list for operator
        op_param_list : list
            register list for operator
        """
        self.operation = {
            "op": op,
            "op_param_list": op_param_list,
            "op_reg_list": op_reg_list,
        }


class Gate_Definition:
    """
    User gate definition.

    Parameters
    ----------
    source : str
        The line to be parsed from the source file.
    filenum : int
        Index of filepath in :class:`T_Sect` filepaths vector.
    linenum : int
        Line number in source file
    param_list : list of str
        Vector of gate parameter names
    ops_raw_list : list of str
        Vecotr of operations as expressed in source code.
    ops_list : list of dict
        Vector of :class:`Gate_Operation` (translated operations)

    Attributes
    ----------
    gate_definition : dict
        Dictionary of the input paramters.
    """

    def __init__(self, source, filenum, linenum, param_list, ops_raw_list, ops_list):
        """
        Instance structures filled in by :class:`QasmTranslator`.
        """
        self.gate_definition = {
            "source": source,
            "filenum": filenum,
            "linenum": linenum,
            "param_list": param_list,
            "ops_raw_list": ops_raw_list,
            "ops_list": ops_list,
        }


# #########################
# Translation unit sections
# #########################


class Source_Body:
    """
    Source code body with file number of source file.

    Parameters
    ----------
    filenum : int
        Index of filepath in :class:`T_Sect` filepaths vector.
    source : str
        The line to be parsed from the source file.

    Attributes
    ----------
    source_body : dict
        Dictionary of the input paramters.
    """

    def __init__(self, filenum, source):
        """
        Instance structures filled in by QasmTranslator
        """
        self.source_body = {"filenum": filenum, "source": source}


class T_Sect:
    """
    Translation overhead section of output AST.

    Parameters
    ----------
    name : str
        The name of the translation unit. Default value is 'main'.

    Attributes
    ----------
    t_sect : dict
        The t_sect of the AST. Will be filled in by :class:`QasmTranslator`.
    """

    def __init__(self, name):
        """Instance structures filled in by :class:`QasmTranslator`."""
        self.t_sect = {
            "name": name,
            "filepaths": [],
            "datetime_start": None,
            "datetime_finish": None,
        }

    def append_filepath(self, filepath):
        """Append to filepaths returning index of latest append"""
        idx = len(self.t_sect["filepaths"])
        self.t_sect["filepaths"].append(filepath)
        return idx


class C_Sect:
    """
    Code section of output AST.

    Attributes
    ----------
    c_sect : list
        The c_sect of the output AST. Vector of :class:`ASTElement`.
    """

    def __init__(self):
        """
        Instance structures filled in by QasmTranslator
        Vector of ASTElement
        """
        self.c_sect = []


class G_Sect:
    """
    User gate definition section of output AST

    Attributes
    ----------
    g_sect : list
        The user gate section of the output AST, to be filled in by
        :class:`QasmTranslator`.

    """

    def __init__(self):
        """Instance structures filled in by QasmTranslator"""
        self.g_sect = []


class S_Sect:
    """
    Source code section of output AST.

    Attributes
    ----------
    s_sect : list
        Vector of :class:`Source_Body` to be filled in by
        :class:`QasmTranslator`.
    """

    def __init__(self):
        """Instance structures filled in by QasmTranslator"""
        self.s_sect = []

    def append(self, sourcebody):
        """
        Appends the source body to the s_sect

        Appends `sourcebody` to :attr: `S_Sect.s_sect`.

        Parameters
        ----------
        sourcebody : dict
            As generated by :attr: `Source_Body.source_body`.
        """
        self.s_sect.append(sourcebody)


# ##################################
# Pushable frames for source nesting
# ##################################


class Source_Frame:
    """
    A pushable frame defining the source we are processing

    Parameters
    ----------
    filenum : int
        Index of filepath in :attr: `T_Sect.t_sect` filepaths vector.
    qasmsourcelines : list of str
        List of the lines from the source code (.qasm) file.

    Attributes
    ----------
    linenum : int
        The index of a line from the source code file.
    """

    def __init__(self, filenum, qasmsourcelines):
        """
        Init counter to 0
        """
        self.filenum = filenum
        self.qasmsourcelines = qasmsourcelines
        self.linenum = 0

    def next(self):
        """Return next source line and increment counter"""
        source = None
        if self.linenum < len(self.qasmsourcelines):
            source = self.qasmsourcelines[self.linenum]
            self.linenum += 1
        return source

    def nth_qasmline(self, n):
        """Return nth qasm source line"""
        return self.qasmsourcelines[n] if n < len(self.qasmsourcelines) else None


class Source_Frame_Stack:
    """
    Stack of source frames so we can nest into include files

    Attributes
    ----------
    frames : list
        The stack.
    """

    def __init__(self):
        """Create the stack"""
        self.frames = []

    def push(self, filenum, qasmsourcelines):
        """
        Create and push a frame

        Parameters
        ----------
        filenum : int
            Index of filepath in :attr: `T_Sect.t_sect` filepaths vector.
        qasmsourcelines : list of str
            List of the lines from the source code (.qasm) file.
        """
        self.frames.append(Source_Frame(filenum, qasmsourcelines))

    def pop(self):
        """Lose top frame"""
        self.frames.pop()

    def tos(self):
        """Return top frame"""
        return self.frames[-1]

    def next(self):
        """Return next line in source from top frame or None"""
        return self.filenum(), self.linenum(), self.tos().next()

    def depth(self):
        """Depth of stack"""
        return len(self.frames)

    def linenum(self):
        """Return linenum of current tos"""
        return self.tos().linenum

    def filenum(self):
        """Return linenum of current tos"""
        return self.tos().filenum

    def nth_qasmline(self, n):
        return self.tos().nth_qasmline(n)


# ##############
# The Translator
# ##############


class QasmTranslator:
    """
    Translation of a Qasm unit.

    Does not read in from file, expects code handed to it.

    Parameters
    ----------
    qasmsourcelines : list of str
        List of the lines from the source code (.qasm) file.
    name : str
        User-defined name for translation unit.
    filepath : str or None
        The filepath to the source code file (informational only). Default
        value is None.
    no_unknown : bool
        True if an unknown element is encountered. Default value is False.
    save_pgm_source : bool
        True if program source should be embedded in output. Default value is
        False
    save_element_source : bool
        True if element source should be embedded in output. Default value is
        False.
    save_gate_source : bool
        True if user gate source should be embedded in output. Default value
        is False.
    save_gate_decls : bool
        True if gate declaration should be noted in c_sect. Default value is
        False.
    include_path : str
        Path to the directory containing the .inc file to be included with the
        OPENQASM 2.0 include statement (typically qelib1.inc). Default value
        is '.' (current working directory).

    Attributes
    ----------
    t_sect : :class:`T_Sect`
    c_sect : :class:`C_Sect`
    g_sect : :class:`G_Sect`
    s_sect : :class:`S_Sect`
    translation : dict
        Dictionary containing all of the sections (t_sect, c_sect, etc).
    source_frame_stack : :class:`Source_Frame_Stack`
    """

    def __init__(
        self,
        qasmsourcelines,
        name="main",
        filepath=None,
        no_unknown=False,
        save_pgm_source=False,
        save_element_source=False,
        save_gate_source=False,
        show_gate_decls=False,
        include_path=".",
    ):
        """
        Init from source lines in an array.
        Does not read in from file, expects code handed to it.
        qasmsourcelines = the source code
        name = user-defined name for translation unit
        no_unknown = True if raises on unknown element
        filepath = source code filepath (informational only)
        save_pgm_source = True if program source should be embedded in output
        save_element_source = True if element source should be embedded in output
        save_gate_source = True if user gate source should be embedded in output
        show_gate_decls = True if gate declaration should be noted in c_sect
        include_path is path for include file search
        """

        # Control factors
        self.no_unknown = no_unknown
        self.save_pgm_source = save_pgm_source
        self.save_element_source = save_element_source
        self.save_gate_source = save_gate_source
        self.show_gate_decls = show_gate_decls
        self.include_path = include_path

        # Init sections
        self.t_sect = T_Sect(name)
        self.c_sect = C_Sect()
        self.g_sect = G_Sect()

        if save_pgm_source is None:
            self.s_sect = None
        else:
            self.s_sect = S_Sect()

        self.translation = {
            "t_sect": self.t_sect.t_sect,
            "c_sect": self.c_sect.c_sect,
            "g_sect": self.g_sect.g_sect,
            "s_sect": self.s_sect.s_sect,
        }

        # Prepare to process initial source
        self.source_frame_stack = Source_Frame_Stack()
        self.push_source(filepath, qasmsourcelines)

    @staticmethod
    def fromFileHandle(
        file_handle,
        name="main",
        filepath=None,
        datetime=None,
        no_unknown=False,
        save_pgm_source=False,
        save_element_source=False,
        save_gate_source=False,
        show_gate_decls=False,
        include_path=".",
    ):
        """
        Instance QasmTranslator from a file handle reading in all lines.

        Does not close file handle.

        Parameters
        ----------
        qasmsourcelines : list of str
            List of the lines from the source code (.qasm) file.
        name : str
            User-defined name for translation unit.
        filepath : str or None
            The filepath to the source code file (informational only). Default
            value is None.
        no_unknown : bool
            True if an unknown element is encountered. Default value is False.
        save_pgm_source : bool
            True if program source should be embedded in output. Default value is
            False
        save_element_source : bool
            True if element source should be embedded in output. Default value is
            False.
        save_gate_source : bool
            True if user gate source should be embedded in output. Default value
            is False.
        save_gate_decls : bool
            True if gate declaration should be noted in c_sect. Default value is
            False.
        include_path : str
            Path to include in file search. Default value is '.' (current
            working directory).

        Returns
        -------
        qt : :class:`QasmTranslator`
        """
        qasmsourcelines = []
        for line in file_handle:
            qasmsourcelines.append(line.strip())
        qt = QasmTranslator(
            qasmsourcelines,
            name=name,
            filepath=filepath,
            no_unknown=no_unknown,
            save_pgm_source=save_pgm_source,
            save_element_source=save_element_source,
            save_gate_source=save_gate_source,
            show_gate_decls=show_gate_decls,
            include_path=include_path,
        )
        return qt

    @staticmethod
    def fromFile(
        filepath,
        name="main",
        no_unknown=False,
        save_pgm_source=False,
        save_element_source=False,
        save_gate_source=False,
        show_gate_decls=False,
        include_path=".",
    ):
        """
        Create parser given filepath.

        Creates :class:`QasmTranslator` given filepath. Opens file indicated
        by `filepath`, reads in all lines and closes file.

        filepath : str or None
            The filepath to the source code file (informational only). Default
            value is None.
        name : str
            User-defined name for translation unit.
        no_unknown : bool
            True if an unknown element is encountered. Default value is False.
        save_pgm_source : bool
            True if program source should be embedded in output. Default value is
            False
        save_element_source : bool
            True if element source should be embedded in output. Default value is
            False.
        save_gate_source : bool
            True if user gate source should be embedded in output. Default value
            is False.
        save_gate_decls : bool
            True if gate declaration should be noted in c_sect. Default value is
            False.
        include_path : str
            Path to include in file search. Default value is '.' (current
            working directory).

        Returns
        -------
        qt : :class:`QasmTranslator`
        """
        if not os.path.exists(filepath) or not os.access(filepath, os.R_OK):
            raise Qasm_Cannot_Read_File_Exception(None, None, None, None, filepath)
        qasmsourcelines = []
        file_handle = open(filepath, "r")
        for line in file_handle:
            qasmsourcelines.append(line.strip())
        file_handle.close()
        qt = QasmTranslator(
            qasmsourcelines,
            name=name,
            filepath=filepath,
            no_unknown=no_unknown,
            save_pgm_source=save_pgm_source,
            save_element_source=save_element_source,
            save_gate_source=save_gate_source,
            show_gate_decls=show_gate_decls,
            include_path=include_path,
        )
        return qt

    def push_source(self, filepath, qasmsourcelines):
        """Add filepath, push source frame stack, and save source if wanted"""
        filenum = self.t_sect.append_filepath(filepath)
        self.source_frame_stack.push(filenum, qasmsourcelines)
        if self.save_pgm_source:
            self.s_sect.append(Source_Body(filenum, qasmsourcelines).source_body)

    def filenum(self):
        """Return the current filenum"""
        return self.source_frame_stack.filenum()

    def linenum(self):
        """Return the current linenum"""
        return self.source_frame_stack.linenum()

    def nth_qasmline(self, n):
        """Return nth line in current source or None"""
        return self.source_frame_stack.nth_qasmline(n)

    def find_include(self, filepath):
        """
        Search include path for filepath
        Return completed filepath if found else None
        """
        found = None
        include_dirs = self.include_path.split(os.pathsep)
        for idir in include_dirs:
            ipath = idir + os.path.sep + filepath
            if os.path.exists(ipath):
                found = ipath
                break
        return found

    def push_include(self, filepath):
        """Open an include file, read it, close it, push source"""
        found = self.find_include(filepath)
        if not found:
            raise Qasm_Cannot_Find_File_Exception(
                self.filenum(),
                self.get_nth_filepath(self.filenum()),
                self.linenum() - 1,
                self.nth_qasmline(self.linenum() - 1),
                filepath,
            )
        filepath = found
        if not os.access(filepath, os.R_OK):
            raise Qasm_Cannot_Read_File_Exception(
                self.filenum(),
                self.get_nth_filepath(self.filenum()),
                self.linenum() - 1,
                self.nth_qasmline(self.linenum() - 1),
                filepath,
            )
        qasmsourcelines = []
        file_handle = open(filepath, "r")
        for line in file_handle:
            qasmsourcelines.append(line.strip())
        file_handle.close()
        self.push_source(filepath, qasmsourcelines)

    def append_ast(self, ast):
        """
        Internal routine to append to the AST
        """
        self.translation["c_sect"].append(ast)

    def append_user_gate(self, user_gate):
        """Append a user gate definition to the user_gates output list"""
        self.translation["g_sect"].append(user_gate)

    def user_gate_definition(self, filenum, linenum, txt):
        """Internal routine to parse and append a user gate definition"""
        txt = txt.strip()
        gate_decl = QTRegEx.GATE_DECL.match(txt)
        gate_name = gate_decl.group(1)
        possible_gate_params = gate_name.split("(")
        gate_param_list = None
        if len(possible_gate_params) > 1:
            gate_param_list = possible_gate_params[1].strip(") ").split(",")
        gate_regs = gate_decl.group(2).strip(
            "{}"
        )  # watch out for no space before '{' !!
        gate_reg_list = gate_regs.split(",")
        gate_ops = QTRegEx.GATE_OPS.match(txt)
        gate_ops_raw_list = QTRegEx.GATE_OPS_LIST.findall(gate_ops.group(1))
        gate_ops_list = []
        for op_raw in gate_ops_raw_list:
            op_raw = op_raw.strip()
            op = QTRegEx.GATE_OP.match(op_raw).group(1)
            op_params = QTRegEx.GATE_OP_PARAMS.match(op_raw)
            if op_params:
                op_param_list = op_params.group(1).split(",")
            else:
                op_param_list = None
            op_regs = QTRegEx.GATE_OP_REGS.match(op_raw)
            if op_regs:
                op_reg_list = op_regs.group(1).split(",")
            else:
                op_reg_list = None
            gate_ops_list.append(
                {"op": op, "op_param_list": op_param_list, "op_reg_list": op_reg_list}
            )
        gate = {
            "source": txt if self.save_gate_source else None,
            "filenum": filenum,
            "linenum": linenum,
            "gate_name": gate_name,
            "gate_param_list": gate_param_list,
            "gate_ops_raw_list": gate_ops_raw_list,
            "gate_ops_list": gate_ops_list,
            "gate_reg_list": gate_reg_list,
        }

        self.append_user_gate(gate)

    def translate(self):
        """
        Translate the qasm source into the desired representation.

        Use `get_translation()` method to retrieve the translated
        source.
        """
        self.get_t_sect()["datetime_start"] = datetime.datetime.now().isoformat()
        seen_noncomment = False
        parsing_gate = False
        gate_def = ""
        gate_start_line = None
        gate_start_linenum = None
        seen_open_curly = False

        while self.source_frame_stack.depth():
            filenum, linenum, line = self.source_frame_stack.next()
            if line is None:
                self.source_frame_stack.pop()
                continue
            astElement = None
            line = line.strip()
            line = line.replace(", ", ",")
            line = line.replace(" ;", ";")
            eolComment = ASTType.ast_eol_comment(line)

            if parsing_gate:
                if not seen_open_curly:
                    x = QTRegEx.START_CURLY.search(line)
                    if not x:
                        raise Qasm_Gate_Missing_Open_Curly_Exception(
                            filenum,
                            self.get_nth_filepath(filenum),
                            linenum,
                            gate_start_line,
                            gate_start_linenum,
                        )
                    else:
                        seen_open_curly = True
                        gate_def = gate_def + line + " "
                        x = QTRegEx.END_CURLY.search(line)
                        if x:
                            self.user_gate_definition(gate_start_linenum, gate_def)
                            parsing_gate = False
                            gate_def = ""
                            gate_start_line = None
                            gate_start_linenum = None
                            seen_open_curly = False
                else:
                    gate_def = gate_def + line + " "
                    x = QTRegEx.END_CURLY.search(line)
                    if x:
                        self.user_gate_definition(filenum, gate_start_linenum, gate_def)
                        parsing_gate = False
                        gate_def = ""
                        gate_start_line = None
                        gate_start_linenum = None
                        seen_open_curly = False

                continue

            astElement = ASTElementUnknown(filenum, linenum, line)
            astType = ASTType.astType(line)
            if astType == ASTType.BLANK:
                continue
            if not seen_noncomment and astType != ASTType.COMMENT:
                if astType == ASTType.DECLARATION_QASM_2_0:
                    seen_noncomment = True
                else:
                    raise Qasm_Declaration_Absent_Exception(
                        filenum, self.get_nth_filepath(filenum), linenum, line
                    )

            # Now step thru types
            if astType == ASTType.COMMENT:
                astElement = ASTElementComment(
                    filenum,
                    linenum,
                    line,
                    self.save_element_source,
                    eol_comment=eolComment,
                )
            elif astType == ASTType.DECLARATION_QASM_2_0:
                astElement = ASTElementDeclarationQasm2_0(
                    filenum,
                    linenum,
                    line,
                    self.save_element_source,
                    eol_comment=eolComment,
                )
            if astType == ASTType.INCLUDE:
                astElement = ASTElementInclude(
                    filenum,
                    linenum,
                    line,
                    self.save_element_source,
                    eol_comment=eolComment,
                )
                self.push_include(astElement.out()["include"])
            elif astType == ASTType.CTL_2:
                astElement = ASTElementCtl2(
                    filenum,
                    linenum,
                    line,
                    self.save_element_source,
                    eol_comment=eolComment,
                )
            elif astType == ASTType.QREG:
                astElement = ASTElementQReg(
                    filenum,
                    linenum,
                    line,
                    self.save_element_source,
                    eol_comment=eolComment,
                )
            elif astType == ASTType.CREG:
                astElement = ASTElementCReg(
                    filenum,
                    linenum,
                    line,
                    self.save_element_source,
                    eol_comment=eolComment,
                )
            elif astType == ASTType.MEASURE:
                astElement = ASTElementMeasure(
                    filenum,
                    linenum,
                    line,
                    self.save_element_source,
                    eol_comment=eolComment,
                )
            elif astType == ASTType.BARRIER:
                astElement = ASTElementBarrier(
                    filenum,
                    linenum,
                    line,
                    self.save_element_source,
                    eol_comment=eolComment,
                )

            elif astType == ASTType.GATE:
                if self.show_gate_decls:
                    astElement = ASTElementGateDefinitionPlaceholder(
                        filenum,
                        linenum,
                        line,
                        self.save_element_source,
                        eol_comment=eolComment,
                    )
                    self.append_ast(astElement.out())
                parsing_gate = True
                gate_start_line = line
                gate_start_linenum = linenum
                gate_def = line + " "
                x = QTRegEx.START_CURLY.search(line)
                if x:
                    seen_open_curly = True
                    x = QTRegEx.END_CURLY.search(line)
                    if x:
                        self.user_gate_definition(filenum, gate_start_linenum, gate_def)
                        parsing_gate = False
                        gate_def = ""
                        gate_start_line = None
                        gate_start_linenum = None
                        seen_open_curly = False
                continue

            elif astType == ASTType.OP:
                astElement = ASTElementOp(
                    filenum,
                    linenum,
                    line,
                    self.save_element_source,
                    eol_comment=eolComment,
                )
            if type(astElement) is ASTElementUnknown and self.no_unknown:
                raise Qasm_Unknown_Element_Exception(
                    filenum, self.get_nth_filepath(filenum), linenum, line
                )
            self.append_ast(astElement.out())

        if parsing_gate:
            raise Qasm_Incomplete_Gate_Exception(
                filenum,
                self.get_nth_filepath(filenum),
                linenum,
                gate_start_line,
                gate_start_linenum,
            )

        self.t_sect.t_sect["datetime_finish"] = datetime.datetime.now().isoformat()

    def get_translation(self):
        """Retrieve translation created by `translate()` method"""
        return self.translation

    # #########################
    # Access Methods for Output
    # #########################

    def get_t_sect(self):
        """Return translation unit section"""
        return self.translation["t_sect"]

    def get_c_sect(self):
        """Return code section"""
        return self.translation["c_sect"]

    def get_g_sect(self):
        """Return gate decl section"""
        return self.translation["g_sect"]

    def get_s_sect(self):
        """Return source section"""
        return self.translation["s_sect"]

    def get_filepaths(self):
        """Retrieve filepaths"""
        return self.get_t_sect()["filepaths"]

    def get_nth_filepath(self, n):
        """Retrieve nth filepath entry"""
        return self.get_filepaths()[n] if n < len(self.get_filepaths()) else None

    def get_datetime_start(self):
        """Retrieve datetime from translation created by `translate()` method"""
        return self.get_t_sect()["datetime_start"]

    def get_datetime_finish(self):
        """Retrieve datetime from translation created by `translate()` method"""
        return self.get_t_sect()["datetime_finish"]

    def get_nth_ast(self, n):
        """Retrieve nth element in c_sect from translation created by `translate()` method"""
        return self.get_c_sect()[n]

    def get_nth_ast_type(self, n):
        """
        Retrieve AST type of nth c_sect element
        from translation created by `translate()` method
        """
        return self.get_nth_ast(n)["type"]

    def get_nth_ast_source(self, n):
        """
        Retrieve source code of nth AST element
        from translation created by `translate()` method
        """
        return self.get_nth_ast(n)["source"]

    def get_source(self, filenum):
        """
        Retrieve original source code of qasm file
        number n that was translated by `translate()` method.
        """
        return self.get_s_sect()[filenum] if filenum < len(self.get_s_sect()) else None

    def get_nth_user_gate(self, index):
        """
        Get nth user_gate definition
        from translation created by `translate()` method
        """
        return self.get_g_sect()[index]


# ##########
# Exceptions
# ##########


class Qasm_Exception(Exception):
    """Base class for Qasm exceptions"""

    def __init__(self, filenum, filename, linenum, line):
        super(Qasm_Exception, self).__init__()
        self.filenum = filenum
        self.filename = filename
        self.linenum = linenum
        self.line = line
        self.message = "Qasm_Exception"
        self.errcode = 10

    def errpacket(self):
        ex = {
            "message": self.message,
            "filenum": self.filenum,
            "filename": self.filename,
            "linenum": self.linenum,
            "line": self.line,
            "errcode": self.errcode,
        }
        return ex


class Qasm_Declaration_Absent_Exception(Qasm_Exception):
    """QASM2.0 Declaration not first non-blank non-comment line"""

    def __init__(self, filenum, filename, linenum, line):
        super(Qasm_Declaration_Absent_Exception, self).__init__(
            filenum, filename, linenum, line
        )
        self.message = "QASM2.0 Declaration not first non-blank non-comment line"
        self.errcode = 20


class Qasm_Unknown_Element_Exception(Qasm_Exception):
    """Unknown element"""

    def __init__(self, filenum, filename, linenum, line):
        super(Qasm_Unknown_Element_Exception, self).__init__(
            filenum, filename, linenum, line
        )
        self.message = "Unknown element"
        self.errcode = 30


class Qasm_Incomplete_Gate_Exception(Qasm_Exception):
    """Gate definition incomplete"""

    def __init__(self, filenum, filename, linenum, line, start_linenum):
        super(Qasm_Incomplete_Gate_Exception, self).__init__(
            filenum, filename, linenum, line
        )
        self.message = "Gate definition incomplete"
        self.errcode = 40
        self.start_linenum = start_linenum

    def errpacket(self):
        ex = {
            "message": self.message,
            "filenum": self.filenum,
            "filename": self.filename,
            "linenum": self.linenum,
            "line": self.line,
            "start_linenum": self.start_linenum,
            "errcode": self.errcode,
        }
        return ex


class Qasm_Gate_Missing_Open_Curly_Exception(Qasm_Exception):
    """Gate definition incomplete"""

    def __init__(self, filenum, filename, linenum, line, start_linenum):
        super(Qasm_Gate_Missing_Open_Curly_Exception, self).__init__(
            filenum, filename, linenum, line
        )
        self.message = "Gate definition missing open curly brace"
        self.errcode = 45
        self.start_linenum = start_linenum

    def errpacket(self):
        ex = {
            "message": self.message,
            "filenum": self.filenum,
            "filename": self.filename,
            "linenum": self.linenum,
            "line": self.line,
            "start_linenum": self.start_linenum,
            "errcode": self.errcode,
        }
        return ex


class Qasm_Cannot_Find_File_Exception(Qasm_Exception):
    """File read error"""

    def __init__(self, filenum, filename, linenum, line, filepath):
        super(Qasm_Cannot_Find_File_Exception, self).__init__(
            filenum, filename, linenum, line
        )
        self.message = "Cannot find file."
        self.errcode = 50
        self.filepath = filepath

    def errpacket(self):
        ex = {
            "message": self.message,
            "filenum": self.filenum,
            "filename": self.filename,
            "linenum": self.linenum,
            "line": self.line,
            "filepath": self.filepath,
            "errcode": self.errcode,
        }
        return ex


class Qasm_Cannot_Read_File_Exception(Qasm_Exception):
    """File read error"""

    def __init__(self, filenum, filename, linenum, line, filepath):
        super(Qasm_Cannot_Read_File_Exception, self).__init__(
            filenum, filename, linenum, line
        )
        self.message = "Cannot read file."
        self.errcode = 55
        self.filepath = filepath

    def errpacket(self):
        ex = {
            "message": self.message,
            "filenum": self.filenum,
            "filename": self.filename,
            "linenum": self.linenum,
            "line": self.line,
            "filepath": self.filepath,
            "errcode": self.errcode,
        }
        return ex


def qasm2ast(
    filepath,
    name="main",
    save_pgm_source=False,
    save_element_source=False,
    save_gate_source=False,
    show_gate_decls=False,
    include_path=".",
):
    """
    Parses a .qasm source file to an abstract syntax tree.

    Parameters
    ----------
    filepaths : list
        The filepaths to the QASM code.
    name : str, optional
        A name for the translation unit (QasmTranslator. The default is 'main'.
    save_pgm_source : bool, optional
        Whether to include program source in AST. The default is False.
    save_element_source : bool, optional
        Whether to include element (typically a line of QASM code) in AST. The
        default is False.
    save_gate_source : bool, optional
        Whether to include gate source in AST output. The default is False.
    show_gate_decls : bool, optional
        Whether to include gate gate declerations in AST output. The default is
        False.
    include_path : str, optional
        Path to include in file search. Default value is '.' (current
        working directory).

    Returns
    -------
    ast : dict of dicts
        An abstract syntax tree (AST) parsed from the QASM
        code given by the filename from filepath. The filename is the key
        for the dictionary.
    """

    def _handle_error(err, erring_filepath):
        """Print out exception packet"""
        x = err.errpacket()
        raise Exception(
            "Error when parsing file" + erring_filepath + ". Error packet: ", x
        )

    qt = None
    try:
        qt = QasmTranslator.fromFile(
            filepath,
            name=name,
            no_unknown=False,  # altered from original script
            save_pgm_source=save_pgm_source,
            save_element_source=save_element_source,
            save_gate_source=save_gate_source,
            show_gate_decls=show_gate_decls,
            include_path=include_path,
        )
        qt.translate()

    except Qasm_Exception as ex:
        _handle_error(ex, filepath)

    translated_ast = qt.get_translation()
    return translated_ast


# =============================================================================
#     filename = os.path.basename(filepath)
#     asts[filename] = translated_ast
# =============================================================================

# fin



"""
Tools for partitioning a monolithic quantum circuit between QPUs.
"""

import itertools
import math

from netsquid.components import instructions as instr

from dqc_simulator.hardware.quantum_processors import QPU


def bisect_circuit(dqc_circuit, comm_qubits_per_node=2):
    """
    Bisects a circuit into two QPUs, node_0 and node_1.

    For circuits with odd
    numbers of qubits, node_0 will be given the extra qubit. It also adds,
    communication qubits to each node. It is assumed that all indices of the
    qubits in the input dqc_circuit are intended to be data qubits (so if the
    input circuit is already partitioned in some way, then this function is
    not appropriate to provide an additional bisection.)

    Parameters
    ----------
    dqc_circuit : :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
        A monolithic quantum circuit.
    comm_qubits_per_node : int
        The number of qubits each QPU node has
    """

    # helper functions:
    def _assign_qubit_to_node(qubit_index, qreg_name, node1starting_index):
        starting_index = (
            dqc_circuit.qregs[qreg_name]["starting_index"] + comm_qubits_per_node
        )
        updated_qubit_index = qubit_index + starting_index
        node_name = "node_0"
        if updated_qubit_index >= node1starting_index:
            updated_qubit_index = (
                updated_qubit_index - node1starting_index + comm_qubits_per_node
            )
            node_name = "node_1"
        return updated_qubit_index, node_name

    # main body:
    total_num_qubits = 2 * comm_qubits_per_node
    for qreg_name in dqc_circuit.qregs:
        qreg = dqc_circuit.qregs[qreg_name]
        total_num_qubits = qreg["size"] + total_num_qubits
    node_0_size = math.ceil(total_num_qubits / 2)  # ceiling division
    node_1_size = total_num_qubits - node_0_size
    node1starting_index = node_0_size  # as indexing starts from zero
    for gate_spec in dqc_circuit.ops:
        qubit_index1 = gate_spec[1]
        qreg_name1 = gate_spec[2]
        updated_qubit_index, node_name = _assign_qubit_to_node(
            qubit_index1, qreg_name1, node1starting_index
        )
        gate_spec[1] = updated_qubit_index
        gate_spec[2] = node_name
        if len(gate_spec) >= 5:
            qubit_index2 = gate_spec[3]
            qreg_name2 = gate_spec[4]
            updated_qubit_index, node_name = _assign_qubit_to_node(
                qubit_index2, qreg_name2, node1starting_index
            )
            gate_spec[3] = updated_qubit_index
            gate_spec[4] = node_name

    dqc_circuit.node_sizes = {"node_0": node_0_size, "node_1": node_1_size}
    dqc_circuit.circuit_type = "partitioned"


def first_come_first_served_qubits_to_qpus(gate_tuples, qpu_nodes):
    """
    Allocates qubits from a monolithic circuit to different QPUs as evenly as
    possible.

    The extra remainder qubits are
    allocated one by one to each QPU node until all are allocated (meaning some
    QPU nodes won't get an extra qubit.)

    Parameters
    ----------
    gate_tuples : list of tuples
        The gates describing a monolithic quantum circuit
    qpu_nodes : list of `~netsquid.node.nodes.Node`s
        The QPU nodes.

    Raises
    ------
    ValueError
        If first instruction in gate tuples is not an
        `~netsquid.components.instructions.Instruction`

    Returns
    -------
    old_to_new_lookup : dict
        Lookup table from the old qubit indices in the orignal monolithic
        circuit to the index and QPU name that qubit is assigned to.
    proc_qubit_allocation4each_qpu : dict
        The processing qubits that need to be initialised on each QPU. The
        keys should be the name of the QPU node and the values should be
        lists of indices.
    """
    num_qpu_nodes = len(qpu_nodes)
    # assuming that first thing that is done is to initialise all data qubits
    init_gate_tuple = gate_tuples[0]
    if init_gate_tuple[0] is not instr.INSTR_INIT:
        raise ValueError(
            "first instruction in gate_tuples must initialise "
            "the processing qubits. The first instruction, "
            f"{init_gate_tuple}, does not do this"
        )
    original_qubit_indices = init_gate_tuple[1]  # num qubits in original circuit
    # defining total number of processing qubits
    total_num_proc_qubits = len(original_qubit_indices)
    proc_qubits_per_qpu = total_num_proc_qubits // num_qpu_nodes
    remainder = len(original_qubit_indices) % num_qpu_nodes
    # initialising values to be filled in by loop
    old_to_new_lookup = {}
    starting_index = 0
    proc_qubit_allocation4each_qpu = {}
    for qpu_node in qpu_nodes:
        end_index = starting_index + proc_qubits_per_qpu
        if remainder > 0:  # if num_qubits/num_qpu has remainder not yet allocated:
            # allocate one qubit from the remainder to this qpu
            end_index = end_index + 1
            remainder = remainder - 1
        # assigning qubits to the qpu
        old_indices_assigned_to_qpu = [ii for ii in range(starting_index, end_index)]
        # creating lookup table to navigate from the qubit indices assigned to
        # each QPU the new indices wrt the QPU those qubits are assigned to
        old2new = {
            old_indices_assigned_to_qpu[ii]: (
                qpu_node.qmemory.processing_qubit_positions[0] + ii,
                qpu_node.name,
            )
            for ii in range(end_index - starting_index)
        }
        proc_qubit_allocation4each_qpu[qpu_node.name] = [
            qpu_node.qmemory.processing_qubit_positions[0] + ii
            for ii in range(end_index - starting_index)
        ]
        # TO DO: maybe combine above two lines into one for loop
        starting_index = end_index
        old_to_new_lookup.update(old2new)
    return old_to_new_lookup, proc_qubit_allocation4each_qpu


def network2qpu_nodes(network):
    """
    Get list of QPU nodes from a network.

    Parameters
    ----------
    network : :class:`~netsquid.nodes.network.Network`
        The quantum network for a quantum data centre.

    Returns
    -------
    qpu_nodes : list of :class:`~netsquid.nodes.node.Node`
        The nodes with a QPU. These are intended to contribute directly to a
        computation rather than performing an administrative function.
    """
    qpu_nodes = []
    for node in network.nodes.values():
        if isinstance(node.qmemory, QPU):
            qpu_nodes.append(node)
    return qpu_nodes


# Can potentially convert the following to a function called partition gate
# tuples which accepts a function to create the old_to_new_lookup as an
# input
def partition_gate_tuples(
    gate_tuples, network, scheme, old_to_new_lookup, proc_qubit_allocation4each_qpu
):
    """


    Parameters
    ----------
    gate_tuples : list of tuples
        The gate_tuples for a monolithic circuit. These can have the form
        (instruction, qubit_index_ii, qubit_index_jj, ...) and/or
        (instruction, qubit_index_ii, node_name, qubit_index_jj, node_name, ...)
        where in the latter case node_name will be a constant (the name of the
        monolithic quantum processor the circuit was originally designed for)
    network : :class:`~netsquid.nodes.network.Network`
        The quantum network for a quantum data centre.
    scheme : str
        The scheme to use for remote gates. Allowed values are 'cat', '1tp',
        '2tp', and 'tp_safe'
    old_to_new_lookup : dict of tuples
        A lookup table with the keys being the
        old indices from a monolithic quantum circuit and the values being
        tuples containing the new index and QPU node name that the
        corresponding old index is assigned to.
    proc_qubit_allocation4each_qpu : dict
        The processing qubits that need to be initialised on each QPU. The
        keys should be the name of the QPU node and the values should be
        lists of indices.

    Returns
    -------
    partitioned_gate_tuples : list of tuples
        The gate tuples for a distributed quantum circuit in the at the
        gate-level. This means the circuit is ready to be compiled by a
        compiler that further decomposes remote gates into communication
        primitives:
            1) single-qubit gate: (gate_instr, qubit, node_name)
            2) two-qubit gate: (list of instructions or gate_instruction or
                                instruction_tuple if local,
                                qubit0, node0_name, qubit1, node1_name, scheme if remote)
                                #can later extend this to multi-qubit gates
                                #keeping scheme as last element
                                list of instructions: list
                                    list of same form as partitioned_gates containing
                                    the local gates to be conducted as part of the
                                    remote gate. Ie, for remote-cnot this would
                                    contain the cnot (but not the gates used for
                                     bsm or correction).Note if this is given as
                                    empty list and scheme = "tp" then it will
                                    just do a teleportation
                                gate_instruction : :class:`~netsquid.components.instructions.Instruction`
                                    The gate instruction for the target gate
                                instruction_tuple : tuple
                                    Tuple of form (gate_instruction, op), where
                                    op is the operatution used to perform the
                                    gate. This form is useful if you want to give
                                    several gates the same
                                    `netsquid.components.qprocessor.PhysicalInstruction`
                                qubit{ii}: int
                                    The qubit index
                                node_name: str
                                The name of the relevant node
                                scheme: str, optional
                                    Only needed for remote gates
    """

    def _split_up_initial_init_instruction(gate_tuples, proc_qubit_allocation4each_qpu):
        """
        Distributes an intialisation instruction at the start of a monolithic quantum
        circuit between differnet QPUs.

        Parameters
        ----------
        gate_tuples : list of tuples
            The gates for a monolithic quantum circuit
        old_to_new_lookup : dict
            Keys should be indices of qubits in monolithic circuit and values should
            be tuples containing the new index for each qubit and the QPU node that
            that qubit now belongs to.

        Returns
        -------
        gate_tuples : list of tuples
            The gate_tuples with the initialisation instruction distributed

        """
        if gate_tuples[0][0] is instr.INSTR_INIT and isinstance(
            gate_tuples[0][1], list
        ):
            # split up the intialisation instruction into an instruction initialisng
            # the relevant qubits on each QPU.
            init_instructions = []
            for node_name in proc_qubit_allocation4each_qpu:
                init_instructions.append(
                    (
                        instr.INSTR_INIT,
                        proc_qubit_allocation4each_qpu[node_name],
                        node_name,
                    )
                )
        else:
            raise ValueError(
                "First gate must be an initialisation. The gate"
                f"{gate_tuples[0]} does not do this."
            )
        return init_instructions

    # assuming that the first gate_tuples is an initialisation instruction
    partitioned_init_instructions = _split_up_initial_init_instruction(
        gate_tuples, proc_qubit_allocation4each_qpu
    )
    partitioned_gate_tuples = []
    # assuming that the first gate_tuples is an initialisation instruction which
    # was dealt with above
    for gate_tuple in gate_tuples[1:]:
        if all(isinstance(entry, int) for entry in gate_tuple[:1]):
            # if all entries after the first are qubit indices:
            step = 1
        else:
            step = 2
        repeating_pattern = list(
            itertools.chain.from_iterable(
                (
                    old_to_new_lookup[gate_tuple[ii]][0],
                    old_to_new_lookup[gate_tuple[ii]][1],
                )
                for ii in range(1, len(gate_tuple), step)
            )
        )
        new_gate = [gate_tuple[0]] + repeating_pattern
        if len(new_gate) > 3 and any(
            new_gate[ii] != new_gate[ii + 2] for ii in range(2, len(new_gate) - 2, 2)
        ):
            # if this is at least a 2 qubit gate and any of the nodes acted on are
            # not the same (ie, if remote gate):
            new_gate = new_gate + [scheme]
        partitioned_gate_tuples.append(tuple(new_gate))
    partitioned_gate_tuples = partitioned_init_instructions + partitioned_gate_tuples
    return partitioned_gate_tuples



# -*- coding: utf-8 -*-
"""
Tools for converting openQASM 2.0 files into a more readily compileable form.

Converts openQASM 2.0 files into a :class:
`~dqc_simulator.software.dqc_circuit.DqcCircuit`.
"""

from netsquid.components import instructions as instr

from dqc_simulator.software.qasm2ast import qasm2ast
from dqc_simulator.software.ast2dqc_circuit import Ast2DqcCircuitTranslator
from dqc_simulator.software.partitioner import bisect_circuit


def preprocess_qasm_to_compilable_bipartitioned(filepath, scheme, include_path="."):
    """
    Converts an openQASM 2.0 file into a bipartioned compileable format.

    Converts an openQASM 2.0 file describing a monolithic quantum circuit
    into a :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit` describing
    a bipartitioned (right down the middle) quantum circuit.

    Parameters
    ----------
    filepath : str
        Filepath to an openQASM 2.0 file describing a monolithic quantum
        circuit.
    scheme : str
        The scheme to use for all inter-QPU (AKA remote) gates. Allowable
        schemes are defined by the compiler.
    include_path : str, optional
        Path to include in file search. The default is '.'.

    Returns
    -------
    dqc_circuit : :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
        A biparitioned distributed quantum circuit.
    """

    ast = qasm2ast(filepath, include_path=include_path)
    dqc_circuit = Ast2DqcCircuitTranslator(ast).ast2dqc_circuit()
    bisect_circuit(dqc_circuit)
    dqc_circuit.add_scheme_to_2_qubit_gates(scheme)
    node_init_commands = []
    for node_name in dqc_circuit.node_sizes:
        new_commands = [
            instr.INSTR_INIT,
            [ii for ii in range(dqc_circuit.node_sizes[node_name])],
            node_name,
        ]
        node_init_commands.append(new_commands)
    dqc_circuit.ops = node_init_commands + dqc_circuit.ops
    return dqc_circuit


def preprocess_qasm_to_compilable_monolithic(filepath, include_path="."):
    """
    Converts an openQASM 2.0 file into an easier to use format.

    Converts an openQASM 2.0 file representing a monolithic quantum circuit
    into :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit` describing
    the same thing.

    Parameters
    ----------
    filepath : str
        The filepath to an openQASM 2.0 file describing a monolithic quantum
        circuit.
    include_path : str, optional
        Path to include in file search. The default is '.'.

    Returns
    -------
    dqc_circuit : :class:`~dqc_simulator.software.dqc_circuit.DqcCircuit`
        A monolithic quantum circuit.

    """
    ast = qasm2ast(filepath, include_path=include_path)
    dqc_circuit = Ast2DqcCircuitTranslator(ast).ast2dqc_circuit()
    # TO DO: add change node name and qubit indices of dqc_circuit.ops
    for ii, gate_spec in enumerate(dqc_circuit.ops):
        qreg1 = dqc_circuit.qregs[gate_spec[2]]
        index1 = gate_spec[1] + qreg1["starting_index"]
        gate_spec[1] = index1
        if len(gate_spec) >= 5:
            qreg2 = dqc_circuit.qregs[gate_spec[4]]
            index2 = gate_spec[3] + qreg2["starting_index"]
            gate_spec[3] = index2
        dqc_circuit.ops[ii] = gate_spec

    dqc_circuit.replace_qreg_names(
        node_0_name="monolithic_qc", node_1_name="monolithic_qc"
    )

    total_num_qubits = 0
    for qreg_name in dqc_circuit.qregs:
        qreg = dqc_circuit.qregs[qreg_name]
        total_num_qubits = qreg["size"] + total_num_qubits
    node_init_cmd = [(instr.INSTR_INIT, list(range(total_num_qubits)), "monolithic_qc")]
    dqc_circuit.ops = node_init_cmd + dqc_circuit.ops
    return dqc_circuit



# -*- coding: utf-8 -*-

# =============================================================================
# Created on Tue May  9 15:16:05 2023
#
# @author: kenny
# =============================================================================
"""Useful circuit identity macros."""

from netsquid.components import instructions as instr

from dqc_simulator.qlib.gates import INSTR_T_DAGGER


def two_control_ibm_toffoli_decomp(
    ctrl_qubit1_index,
    ctrl_node_name1,
    ctrl_qubit2_index,
    ctrl_node_name2,
    target_qubit_index,
    target_node_name,
    scheme="cat",
):
    """
    A macro for the decomposition of a toffoli, or CCX, gate in terms U, CX, H
    and T gates.

    For direct use by protocols in
    :mod: `~dqc_simulator.software.dqc_control`.

    Parameters
    ----------
    ctrl_qubit1_index, ctrl_qubit_2_index : int
        The index of the first and second control qubits of the toffoli gate
    ctrl_node_name1, ctrl_node_name2 : str
        The names of the QPU nodes upon which the first and second control
        qubits of the toffoli gate reside.
    target_qubit_index : int
        The index of the target qubit for the toffoli gate
    target_node_name : str
        DESCRIPTION.
    scheme : str, optional
        The remote gate scheme to use. The default is "cat" (referring to
        cat-comm (AKA telegate).

    Returns
    -------
    sub_ops : list of tuples
        The gates constituting the decomposed toffoli gate.

    Notes
    -----
    This decomposition of the toffoli gate is as specified in the IBM
    Quantum Experience standard header for openQASM 2.0 [1]_.

    References
    ----------
    ..[1] A. W. Cross, L. S. Bishop, J. A. Smolin, and J. M. Gambetta, Open
        Quantum Assembly Language, arXiv:1707.03429 [quant-ph].

    """
    # INSTR_T_dagger defined in my network.py module
    sub_ops = [
        (instr.INSTR_H, target_qubit_index, target_node_name),
        (
            instr.INSTR_CNOT,
            ctrl_qubit2_index,
            ctrl_node_name2,
            target_qubit_index,
            target_node_name,
        ),
        (INSTR_T_DAGGER, target_qubit_index, target_node_name),
        (
            instr.INSTR_CNOT,
            ctrl_qubit1_index,
            ctrl_node_name1,
            target_qubit_index,
            target_node_name,
        ),
        (instr.INSTR_T, target_qubit_index, target_node_name),
        (
            instr.INSTR_CNOT,
            ctrl_qubit2_index,
            ctrl_node_name2,
            target_qubit_index,
            target_node_name,
        ),
        (INSTR_T_DAGGER, target_qubit_index, target_node_name),
        (
            instr.INSTR_CNOT,
            ctrl_qubit1_index,
            ctrl_node_name1,
            target_qubit_index,
            target_node_name,
        ),
        (instr.INSTR_T, ctrl_qubit2_index, ctrl_node_name2),
        (instr.INSTR_T, target_qubit_index, target_node_name),
        (
            instr.INSTR_CNOT,
            ctrl_qubit1_index,
            ctrl_node_name1,
            ctrl_qubit2_index,
            ctrl_node_name2,
        ),
        (instr.INSTR_H, target_qubit_index, target_node_name),
        (instr.INSTR_T, ctrl_qubit1_index, ctrl_node_name1),
        (INSTR_T_DAGGER, ctrl_qubit2_index, ctrl_node_name2),
        (
            instr.INSTR_CNOT,
            ctrl_qubit1_index,
            ctrl_node_name1,
            ctrl_qubit2_index,
            ctrl_node_name2,
        ),
    ]
    if ctrl_node_name1 == ctrl_node_name2 == target_node_name:
        return sub_ops
    else:
        for ii, element in enumerate(sub_ops):
            if len(element) > 3 and element[2] != element[-1]:
                sub_ops[ii] = (*element, scheme)
    return sub_ops


# stabiliser circuits
# -------------------


def stabiliser_measurement(qubits2check, ancilla_to_use, stabiliser_type):
    """
    Specifies circuit for single stabiliser measurement using one ancilla
    qubit.

    Parameters
    ----------
    qubits2check : list or tuple of int
        The qubits (specified by their indices_ to apply the (non-destructive)
        stabiliser measurement to.
    ancilla_to_use : int
        The qubit (index) to use as an ancilla.
    stabiliser_type : str
        Whether to apply an 'x' or a 'z' stabiliser measurement. Allowed values
        are 'x' or 'X' and 'z' or 'Z'

    Returns
    -------
    gate_tuples : list of tuples
        The circuit specification for the stabiliser measurement.

    Notes
    -----
    This method is based on a general one for measuring an operator and is
    discussed in page 473 of Nielsen and Chuang's textbook. [1]_

    This method of stabiliser measurement is somewhat naive and lacks
    fault-tolerance for errors in the measurements themselves.

    References
    ----------
    .. [1] M. Nielsen and I. Chuang, Quantum Computation and Quantum Information, 10th ed. (Cambridge University Press, 2010).
    """
    # in next line 'mono_qc' is placeholder for a node name and can be replaced
    # later
    if (stabiliser_type == "x") or (stabiliser_type == "X"):
        entangling = []
        for qubit_index in qubits2check:
            gates2add = [
                (instr.INSTR_H, qubit_index, "mono_qc"),
                (instr.INSTR_CNOT, qubit_index, "mono_qc", ancilla_to_use, "mono_qc"),
                (instr.INSTR_H, qubit_index, "mono_qc"),
            ]
            entangling = entangling + gates2add
    elif (stabiliser_type == "z") or (stabiliser_type == "Z"):
        entangling = [
            (instr.INSTR_CNOT, qubit_index, "mono_qc", ancilla_to_use, "mono_qc")
            for qubit_index in qubits2check
        ]
    else:
        raise ValueError(
            f"{stabiliser_type} is not an allowed value of "
            'stabiliser_type. Allowed values are: "x", "X", "z", '
            'or "Z". '
        )
    measurement = [(instr.INSTR_MEASURE, ancilla_to_use, "mono_qc", "logging")]
    return entangling + measurement


# =============================================================================
# def steane_code_stabiliser_measurement():
#     """
#     Generate gate tuples for Steane code on a monolithic quantum processor.
#
#     Returns
#     -------
#     None.
#
#     """
#     data_qubit_indices = [ii for ii in range(7)]
#     ancilla_qubit_indices = [ii for ii in range(7, )]
# =============================================================================



"""
Functions and classes for creating specific QPUs.
"""

from netsquid.components import instructions as instr
from netsquid.components.qprocessor import QuantumProcessor, PhysicalInstruction
from netsquid.components.models.qerrormodels import DepolarNoiseModel

from dqc_simulator.hardware.noise_models import (
    AnalyticalDepolarisationModel,
    MeasurementNoiseModel,
    NDimDepolarNoiseModel,
)
from dqc_simulator.qlib.gates import (
    INSTR_ARB_GEN,
    INSTR_CH,
    INSTR_CT,
    INSTR_T_DAGGER,
    INSTR_S_DAGGER,
    INSTR_SINGLE_QUBIT_UNITARY,
    INSTR_TWO_QUBIT_UNITARY,
    INSTR_SINGLE_QUBIT_NEGLIGIBLE_TIME,
    INSTR_TWO_QUBIT_NEGLIGIBLE_TIME,
)


class QPU(QuantumProcessor):
    """


    Parameters
    ----------
    name : TYPE
        DESCRIPTION.
    num_positions : int, optional
        The number of qubits on the memory. The default is 20. Note that
        additional fictitious qubits will also be placed on the memory to
        allow modelling of emitted photons. Therefore, the num_positions
        property will differ from the value specified here.
    mem_noise_models : TYPE, optional
        DESCRIPTION. The default is None. This will be overriden if
        comm_qubit_mem_noise_model or
    phys_instructions : TYPE, optional
        DESCRIPTION. The default is None.
    fallback_to_nonphysical : TYPE, optional
        DESCRIPTION. The default is False.
    properties : TYPE, optional
        DESCRIPTION. The default is None.
    num_comm_qubits : TYPE, optional
        DESCRIPTION. The default is 0.
    comm_qubit_mem_noise_model : :class:`~netsquid.components.models.qerrormodels.QuantumErrorModel`
        The memory noise model for all comm qubits. If this is not None,
        mem_noise_models will be overridden.
    processing_qubit_mem_noise_model : :class:`~netsquid.components.models.qerrormodels.QuantumErrorModel`
        The memory noise model for all processing qubits. If this is not None,
        mem_noise_models will be overridden.

    Attributes
    ----------
    mem_pos_types : list of str
        The type of each memory position. Can be 'comm', 'processing' or
        'photon', reflecting the types of qubits that can occupy the position.
    comm_qubit_positions : list of int
        The memory positions of the communication qubits
    processing_qubit_positions : list of int
        The memory positions of the processing qubits
    photon_positions :list of int
        The fictitious memory positions reserved for facilitating photon
        emission.
    num_real_positions : list of int
        The number of real memory positions (which exludes those reserved for
        modelling photon emission).
    comm_qubits_free : list of int
        The positions of communication qubits which are free to use. Here,
        'free to use' means that they are not hosting quantum information which
        should not be overwritten.
    ebit_ready : bool
        Whether there is an ebit (entangled bit) ready or not. This defaults to
        False and will become True when entanglement is successfully
        distributed between this `QPU` and another.
    """

    def __init__(
        self,
        name,
        num_positions=20,
        mem_noise_models=None,
        phys_instructions=None,
        fallback_to_nonphysical=False,
        properties=None,
        num_comm_qubits=0,
        comm_qubit_mem_noise_model=None,
        processing_qubit_mem_noise_model=None,
    ):
        mem_pos_types, num_photon_spaces = self._get_mem_pos_types_N_num_photons(
            num_positions, num_comm_qubits
        )
        self.mem_pos_types = mem_pos_types
        if comm_qubit_mem_noise_model or processing_qubit_mem_noise_model is not None:
            mem_noise_models = self._get_mem_noise_models(
                mem_pos_types,
                comm_qubit_mem_noise_model,
                processing_qubit_mem_noise_model,
            )
        super().__init__(
            name,
            num_positions=num_positions + num_photon_spaces,
            mem_noise_models=mem_noise_models,
            phys_instructions=phys_instructions,
            mem_pos_types=mem_pos_types,
            fallback_to_nonphysical=fallback_to_nonphysical,
            properties=properties,
        )
        self.comm_qubit_positions = self.get_positions_matching_type("comm")
        self.processing_qubit_positions = self.get_positions_matching_type("processing")
        self.photon_positions = self.get_positions_matching_type("photon")
        self.num_real_positions = self.num_positions - len(self.photon_positions)
        # instantiating the comm_qubits_free attribute to be the same as
        # comm_qubit_positions (initially). The list constructor is important
        # here to make sure that comm-qubits free is a copy of
        # comm_qubit_positions rather than an alias for it.
        self.comm_qubits_free = list(self.comm_qubit_positions)
        self.ebit_ready = False

    def get_positions_matching_type(self, pos_type):
        """
        Gets indices of memory positions matching the given type.

        Parameters
        ----------
        pos_type : str
            The type of memory position. Allowed values are 'comm',
            'processing', and 'photon'.

        Returns
        -------
        positions_matching_pos_type : list of int
            The memory positions with the specified `pos_type`.
        """
        # I could also have used boolean masking with numpy for the following
        # but that method is slower for the number of memory positions that
        # would be feasibly simulable (according to ipython's %timeit)
        positions_matching_pos_type = []
        for ii, mem_pos_type in enumerate(self.mem_pos_types):
            if mem_pos_type == pos_type:
                positions_matching_pos_type.append(ii)
        return positions_matching_pos_type

    def _get_mem_pos_types_N_num_photons(self, num_positions, num_comm_qubits):
        """
        Create a list of memory position types.

        The output of this should be of a form useable as the value for the
        `mem_pos_types` parameter for a
        `~netsquid.components.qprocessor.QuantumProcessor`.

        Parameters
        ----------
        num_positions : int
            The total number of memory positions.
        num_comm_qubits : int
            The number of comm_qubits.

        Returns
        -------
        mem_pos_types : list of str
            The types of memory position. There should be one string for each
            memory position.

        """
        # determining number of fictitious memory positions needed to model photon
        # emission
        num_photon_spaces = num_comm_qubits
        # photons are not included in the next line because num_positions
        num_processing_qubits = num_positions - num_comm_qubits
        mem_pos_types = (
            ["comm"] * num_comm_qubits
            + ["processing"] * num_processing_qubits
            + ["photon"] * num_photon_spaces
        )
        return mem_pos_types, num_photon_spaces

    def _get_mem_noise_models(
        self,
        mem_pos_types,
        comm_qubit_mem_noise_model,
        processing_qubit_mem_noise_model,
    ):
        """
        Gets an ordered list of memory noise models

        Parameters
        ----------
        mem_pos_types : list of str
            The types of memory position. There should be one string for each
            memory position.

        Returns
        -------
        mem_noise_models: list of :class:`~netsquid.components.models.qerrormodels.QuantumErrorModel`
            The noise models to apply to the corresponding memory positions.
        """
        mem_noise_models = []
        for mem_pos_type in mem_pos_types:
            if mem_pos_type == "comm":
                mem_noise_models.append(comm_qubit_mem_noise_model)
            elif mem_pos_type == "processing":
                mem_noise_models.append(processing_qubit_mem_noise_model)
            elif mem_pos_type == "photon":
                mem_noise_models.append(None)  # None because photons are not sitting
                # within the memory. Any noise they
                # have should be modelled using an
                # emission noise model.
            else:
                raise ValueError(
                    f'"{mem_pos_type}" is not of one of the allowed '
                    "values for mem_pos_type for this processor."
                    'The allowed values are: "comm", "data", and '
                    '"photon".'
                )
        return mem_noise_models


def create_processor(
    alpha=1,
    beta=0,
    depolar_rate=0,
    dephase_rate=0,
    num_positions=7,
    mem_pos_types=None,
    name="quantum_processor",
):
    """Factory to create a quite artificial quantum processor.

    Has seven memory positions and the physical instructions necessary
    for remote gates using single and two-qubit local gates.

    Parameters
    ----------
    depolar_rate : float
        Depolarization rate of qubits in memory.
    dephase_rate : float
        Dephasing rate of physical measurement instruction.

    Returns
    -------
    :class:`~netsquid.components.qprocessor.QuantumProcessor`
        A quantum processor to specification.

    Notes
    -----
    The gate durations are too small to be physical. This processor was used
    for cases where gate duration (as long as it was finite to avoid a
    :class:`~netsquid.components.qprocessor.ProcessorBusyError`) did not
    matter.
    """
    # =============================================================================
    #     measure_noise_model = DephaseNoiseModel(dephase_rate=dephase_rate,
    #                                             time_independent=True)
    # =============================================================================
    x_gate_duration = 1
    physical_instructions = [
        PhysicalInstruction(
            instr.INSTR_INIT, duration=3, parallel=False, toplogy=[2, 3, 4, 5, 6]
        ),
        PhysicalInstruction(instr.INSTR_H, duration=1, parallel=True, topology=None),
        PhysicalInstruction(
            instr.INSTR_X, duration=x_gate_duration, parallel=True, topology=None
        ),
        PhysicalInstruction(instr.INSTR_Z, duration=1, parallel=True, topology=None),
        PhysicalInstruction(instr.INSTR_S, duration=1, parallel=True, topology=None),
        PhysicalInstruction(instr.INSTR_CNOT, duration=4, parallel=True, topology=None),
        PhysicalInstruction(INSTR_ARB_GEN(alpha, beta), duration=4, parallel=False),
        PhysicalInstruction(INSTR_CH, duration=4, parallel=True, topology=None),
        PhysicalInstruction(INSTR_CT, duration=4, parallel=True, topology=None),
        PhysicalInstruction(instr.INSTR_CS, duration=4, parallel=True, topology=None),
        PhysicalInstruction(
            instr.INSTR_MEASURE,
            duration=7,
            parallel=False,
            topology=None,
            quantum_noise_model=None,
            apply_q_noise_after=False,
            discard=True,
        ),
        PhysicalInstruction(
            instr.INSTR_DISCARD, duration=3, parallel=False, toplology=[0, 1]
        ),
        PhysicalInstruction(
            instr.INSTR_SWAP, duration=12, parallel=True, topology=None
        ),
        PhysicalInstruction(instr.INSTR_T, duration=1, parallel=True, topology=None),
        PhysicalInstruction(INSTR_T_DAGGER, duration=1, parallel=True, topology=None),
    ]
    memory_noise_model = DepolarNoiseModel(depolar_rate=depolar_rate)
    processor = QuantumProcessor(
        name,
        num_positions=num_positions,
        mem_noise_models=([memory_noise_model] * num_positions),
        phys_instructions=physical_instructions,
        mem_pos_types=mem_pos_types,
    )
    return processor


def _phys_instructions_4_standard_lib_gates_and_convenience_ops(
    cnot_depolar_model,
    single_qubit_gate_noise_model,
    meas_noise_model,
    num_comm_qubits,
    single_qubit_gate_time,
    two_qubit_gate_time,
    measurement_time,
    alpha,
    beta,
):
    """
    Physical instructions for the standard lib gates and some additional
    convenience operators (eg, gates to make an arbitrary state for the start
    of an experiment).

    Parameters
    ----------
    cnot_depolar_model : TYPE
        DESCRIPTION.
    num_comm_qubits : TYPE
        DESCRIPTION.
    single_qubit_gate_time : TYPE
        DESCRIPTION.
    two_qubit_gate_time : TYPE
        DESCRIPTION.
    measurement_time : TYPE
        DESCRIPTION.
    alpha : TYPE
        DESCRIPTION.
    beta : TYPE
        DESCRIPTION.

    Returns
    -------
    physical_instructions : list of :class:`~netsquid.components.qprocessor.PhysicalInstruction`s
    """
    # creating processor for all Nodes
    physical_instructions = [
        PhysicalInstruction(instr.INSTR_INIT, duration=3, parallel=False, toplogy=None),
        PhysicalInstruction(
            instr.INSTR_H,
            duration=single_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=single_qubit_gate_noise_model,
        ),
        PhysicalInstruction(
            instr.INSTR_X,
            duration=single_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=single_qubit_gate_noise_model,
        ),
        PhysicalInstruction(
            instr.INSTR_Z,
            duration=single_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=single_qubit_gate_noise_model,
        ),
        PhysicalInstruction(
            instr.INSTR_S,
            duration=single_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=single_qubit_gate_noise_model,
        ),
        PhysicalInstruction(
            INSTR_S_DAGGER,
            duration=single_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=single_qubit_gate_noise_model,
        ),
        PhysicalInstruction(
            instr.INSTR_T,
            duration=single_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=single_qubit_gate_noise_model,
        ),
        PhysicalInstruction(
            INSTR_T_DAGGER,
            duration=single_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=single_qubit_gate_noise_model,
        ),
        PhysicalInstruction(
            instr.INSTR_CNOT,
            duration=two_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=cnot_depolar_model,
        ),
        PhysicalInstruction(
            instr.INSTR_CZ,
            duration=two_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=cnot_depolar_model,
        ),
        PhysicalInstruction(INSTR_ARB_GEN(alpha, beta), duration=3, parallel=True),
        # duration deliberately negligible
        # to approximate ideality for
        # when not interested in how input
        # state arrived at
        PhysicalInstruction(
            instr.INSTR_MEASURE,
            duration=measurement_time,
            parallel=True,
            topology=None,
            quantum_noise_model=meas_noise_model,
            apply_q_noise_after=False,
            discard=True,
        ),
        PhysicalInstruction(
            instr.INSTR_DISCARD,
            duration=single_qubit_gate_time,
            parallel=False,
            toplology=[ii for ii in range(num_comm_qubits)],
        ),
        # =============================================================================
        #             PhysicalInstruction(instr.INSTR_SWAP, duration=1e-10,
        #                                 parallel=True,
        #                                 topology=None), #duration deliberately negligible
        #                                                 #to approximate ideality for
        #                                                 #TP-safe. MAYBE REVISIT
        # =============================================================================
        PhysicalInstruction(
            INSTR_SINGLE_QUBIT_UNITARY,
            duration=single_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=single_qubit_gate_noise_model,
        ),
        PhysicalInstruction(
            INSTR_TWO_QUBIT_UNITARY,
            duration=two_qubit_gate_time,
            parallel=True,
            topology=None,
            quantum_noise_model=cnot_depolar_model,
        ),
        PhysicalInstruction(
            INSTR_SINGLE_QUBIT_NEGLIGIBLE_TIME, duration=3, parallel=True, topology=None
        ),
        PhysicalInstruction(
            INSTR_TWO_QUBIT_NEGLIGIBLE_TIME, duration=3, parallel=True, topology=None
        ),
    ]
    return physical_instructions


def create_qproc_with_analytical_noise_ionQ_aria_durations_N_standard_lib_gates(
    p_depolar_error_cnot=0,
    comm_qubit_depolar_rate=0,
    proc_qubit_depolar_rate=0,
    single_qubit_gate_time=135 * 10**3,
    two_qubit_gate_time=600 * 10**3,
    measurement_time=600 * 10**4,
    alpha=1,
    beta=0,
    num_positions=20,
    num_comm_qubits=2,
):
    """
    Creates quantum processor loosely based on ionQ Aria but with different
    native gates.

    .. deprecated::
        This function will be removed in later version as its functionality is
        now provided by the more general NoisyQPU class in a style more
        consistent with object-orientated programming

    Parameters
    ----------
    p_depolar_error_cnot : float, optional
        The probability of a depolarisation error after each CNOT gate.
        The default value is 0.
    comm_qubit_depolar_rate : float, optional
        The depolarisation rate for communication qubits.
        The default value is 0.
    proc_qubit_depolar_rate : float
        The depolarisation rate for processing qubits.
        The default value is 0.
    single_qubit_gate_time : TYPE, optional
        DESCRIPTION. The default is 135 * 10**3.
    two_qubit_gate_time : TYPE, optional
        DESCRIPTION. The default is 600 * 10**3.
    measurement_time : TYPE, optional
        DESCRIPTION. The default is 600 * 10**4.
    alpha, beta : float, optional
        Settings for the artificial convenience gate INSTR_ARB_GEN, which
        initialises the state of a qubit as alpha |0> + beta |1>. The default
        is 1 for alpha and 0 for beta.
    num_positions : int, optional
        The number of memory positions. Note that this does not include the
        number of photons (which each occupy a fictitious memory position). The
        default is 20.
    num_comm_qubits : TYPE, optional
        DESCRIPTION. The default is 2.

    Returns
    -------
    qprocessor : :class:`~netsquid.components.qprocessor.QuantumProcessor`

    Notes
    -----
    Uses universal gate sets from Ref. [1]_. Analytical noise is used.
    Gate durations are taken from averages for single and two-qubit gates on
    IonQ's Aria machine.

    References
    ----------
    .. [1] M. Nielsen and I. Chuang, Quantum Computation and Quantum
        Information, 10th ed. (Cambridge University Press, 2010).
    """
    cnot_depolar_model = AnalyticalDepolarisationModel(
        p_error=p_depolar_error_cnot, time_independent=True
    )
    # TO DO: decide whether to replace the following with NetSquid's built-in
    # DepolarNoiseModel which gives the same results when the DM formalism is
    # used and allows other formalisms to also be used.
    comm_qubit_memory_depolar_model = AnalyticalDepolarisationModel(
        comm_qubit_depolar_rate, time_independent=False
    )
    processing_qubit_memory_depolar_model = AnalyticalDepolarisationModel(
        proc_qubit_depolar_rate, time_independent=False
    )
    single_qubit_gate_noise_model = None
    meas_noise_model = None
    # creating processor for all Nodes
    physical_instructions = _phys_instructions_4_standard_lib_gates_and_convenience_ops(
        cnot_depolar_model,
        single_qubit_gate_noise_model,
        meas_noise_model,
        num_comm_qubits,
        single_qubit_gate_time,
        two_qubit_gate_time,
        measurement_time,
        alpha,
        beta,
    )
    qprocessor = QPU(
        "custom_noisy_qprocessor",
        phys_instructions=physical_instructions,
        num_positions=num_positions,
        num_comm_qubits=num_comm_qubits,
        comm_qubit_mem_noise_model=comm_qubit_memory_depolar_model,
        processing_qubit_mem_noise_model=processing_qubit_memory_depolar_model,
    )

    qprocessor.add_composite_instruction(
        instr.INSTR_SWAP,
        [
            (instr.INSTR_CNOT, (0, 1)),
            (instr.INSTR_CNOT, (1, 0)),
            (instr.INSTR_CNOT, (0, 1)),
        ],
        topology=None,
    )
    return qprocessor


def create_noisy_qpu(
    p_depolar_error_cnot=0,
    single_qubit_gate_error_prob=0,
    meas_error_prob=0,
    comm_qubit_depolar_rate=0,
    proc_qubit_depolar_rate=0,
    single_qubit_gate_time=135 * 10**3,
    two_qubit_gate_time=600 * 10**3,
    measurement_time=600 * 10**4,
    alpha=1,
    beta=0,
    num_positions=20,
    num_comm_qubits=2,
):
    """
    Creates quantum processor loosely based on ionQ Aria but with different
    native gates.

    .. deprecated::
        Use NoisyQPU instead.

    Parameters
    ----------
    p_depolar_error_cnot : float, optional
        The probability of a depolarisation error after each CNOT gate.
        The default value is 0.
    single_qubit_gate_error_prob : float
        The probability of a depolarisation error occuring during a single
        qubit gate.
    meas_error_prob : float
        The probability of a bit flip error occuring during measurement, giving
        the opposite result from an ideal measurement.
    comm_qubit_depolar_rate : float, optional
        The depolarisation rate for communication qubits.
        The default value is 0.
    proc_qubit_depolar_rate : float
        The depolarisation rate for processing qubits.
        The default value is 0.
    single_qubit_gate_time : TYPE, optional
        DESCRIPTION. The default is 135 * 10**3.
    two_qubit_gate_time : TYPE, optional
        DESCRIPTION. The default is 600 * 10**3.
    measurement_time : TYPE, optional
        DESCRIPTION. The default is 600 * 10**4.
    alpha, beta : float, optional
        Settings for the artificial convenience gate INSTR_ARB_GEN, which
        initialises the state of a qubit as alpha |0> + beta |1>. The default
        is 1 for alpha and 0 for beta.
    num_positions : TYPE, optional
        DESCRIPTION. The default is 20.
    num_comm_qubits : TYPE, optional
        DESCRIPTION. The default is 2.

    Returns
    -------
    qprocessor : :class:`netsquid.components.qprocessor.QuantumProcessor

    Notes
    -----
    Uses universal gate sets from Ref. [1]_. Numerical noise is used.
    Gate durations are taken from averages for single and two-qubit gates on
    IonQ's Aria machine.

    References
    ----------
    .. [1] M. Nielsen and I. Chuang, Quantum Computation and Quantum
        Information, 10th ed. (Cambridge University Press, 2010).
    """
    cnot_depolar_model = NDimDepolarNoiseModel(
        p_depolar_error_cnot, time_independent=True
    )
    comm_qubit_memory_depolar_model = DepolarNoiseModel(
        comm_qubit_depolar_rate, time_independent=False
    )
    processing_qubit_memory_depolar_model = DepolarNoiseModel(
        proc_qubit_depolar_rate, time_independent=False
    )
    meas_noise_model = MeasurementNoiseModel(meas_error_prob)
    single_qubit_gate_noise_model = DepolarNoiseModel(
        single_qubit_gate_error_prob, time_independent=True
    )
    # creating processor for all Nodes
    physical_instructions = _phys_instructions_4_standard_lib_gates_and_convenience_ops(
        cnot_depolar_model,
        single_qubit_gate_noise_model,
        meas_noise_model,
        num_comm_qubits,
        single_qubit_gate_time,
        two_qubit_gate_time,
        measurement_time,
        alpha,
        beta,
    )
    qprocessor = QPU(
        "custom_noisy_qprocessor",
        phys_instructions=physical_instructions,
        num_positions=num_positions,
        num_comm_qubits=num_comm_qubits,
        comm_qubit_mem_noise_model=comm_qubit_memory_depolar_model,
        processing_qubit_mem_noise_model=processing_qubit_memory_depolar_model,
    )

    qprocessor.add_composite_instruction(
        instr.INSTR_SWAP,
        [
            (instr.INSTR_CNOT, (0, 1)),
            (instr.INSTR_CNOT, (1, 0)),
            (instr.INSTR_CNOT, (0, 1)),
        ],
        topology=None,
    )
    return qprocessor


def create_qproc_with_numerical_noise_ionQ_aria_durations_N_standard_lib_gates(
    p_depolar_error_cnot=0,
    single_qubit_gate_error_prob=0,
    meas_error_prob=0,
    comm_qubit_depolar_rate=0,
    proc_qubit_depolar_rate=0,
    single_qubit_gate_time=135 * 10**3,
    two_qubit_gate_time=600 * 10**3,
    measurement_time=600 * 10**4,
    alpha=1,
    beta=0,
    num_positions=20,
    num_comm_qubits=2,
):
    """
    Creates quantum processor loosely based on ionQ Aria but with different
    native gates.

    .. deprecated::
        Use noisy QPU instead.

    Parameters
    ----------
    p_depolar_error_cnot : float, optional
        The probability of a depolarisation error after each CNOT gate.
        The default value is 0.
    comm_qubit_depolar_rate : float, optional
        The depolarisation rate for communication qubits.
        The default value is 0.
    proc_qubit_depolar_rate : float
        The depolarisation rate for processing qubits.
        The default value is 0.
    single_qubit_gate_time : TYPE, optional
        DESCRIPTION. The default is 135 * 10**3.
    two_qubit_gate_time : TYPE, optional
        DESCRIPTION. The default is 600 * 10**3.
    measurement_time : TYPE, optional
        DESCRIPTION. The default is 600 * 10**4.
    alpha, beta : float, optional
        Settings for the artificial convenience gate INSTR_ARB_GEN, which
        initialises the state of a qubit as alpha |0> + beta |1>. The default
        is 1 for alpha and 0 for beta.
    num_positions : TYPE, optional
        DESCRIPTION. The default is 20.
    num_comm_qubits : TYPE, optional
        DESCRIPTION. The default is 2.

    Returns
    -------
    qprocessor : :class:`netsquid.components.qprocessor.QuantumProcessor

    Notes
    -----
    Uses universal gate sets from Ref. [1]_. Numerical noise is used.
    Gate durations are taken from averages for single and two-qubit gates on
    IonQ's Aria machine.

    References
    ----------
    .. [1] M. Nielsen and I. Chuang, Quantum Computation and Quantum
        Information, 10th ed. (Cambridge University Press, 2010).
    """
    return create_noisy_qpu(
        p_depolar_error_cnot=p_depolar_error_cnot,
        single_qubit_gate_error_prob=single_qubit_gate_error_prob,
        meas_error_prob=meas_error_prob,
        comm_qubit_depolar_rate=comm_qubit_depolar_rate,
        proc_qubit_depolar_rate=proc_qubit_depolar_rate,
        single_qubit_gate_time=single_qubit_gate_time,
        two_qubit_gate_time=two_qubit_gate_time,
        measurement_time=measurement_time,
        alpha=alpha,
        beta=beta,
        num_positions=num_positions,
        num_comm_qubits=num_comm_qubits,
    )


# =============================================================================
# def create_qproc4oxford_dqc_proof_of_principle(
#                                          p_depolar_error_cnot=0,
#                                          comm_qubit_depolar_rate=0,
#                                          proc_qubit_depolar_rate=0,
#                                          single_qubit_gate_time=135 * 10**3,
#                                          two_qubit_gate_time=600 * 10**3,
#                                          measurement_time=600 * 10**4,
#                                          alpha=1, beta=0,
#                                          num_positions=20,
#                                          num_comm_qubits=2):
#     """
#     Creates a QPU designed to be compared to those in _[1].
#
#     References
#     ----------
#     .. [1] D. Main, P. Drmota, D. P. Nadlinger, E. M. Ainley, A. Agrawal,
#         B. C. Nichol, et al., Distributed Quantum Computing across an Optical
#         Network Link, arXiv:2407.00835.
#     """
#     cnot_depolar_model = AnalyticalDepolarisationModel(
#                                 p_error=p_depolar_error_cnot,
#                                 time_independent=True)
#     comm_qubit_memory_depolar_model = DepolarNoiseModel(
#                                          comm_qubit_depolar_rate,
#                                          time_independent=False)
#     processing_qubit_memory_depolar_model = DepolarNoiseModel(
#                                                 proc_qubit_depolar_rate,
#                                                 time_independent=False)
#     physical_instructions = [] #FILL THIS LIST IN
# =============================================================================


class NoisyQPU(QPU):
    """
    Creates quantum processor loosely based on ionQ Aria but with different
    native gates.

    Parameters
    ----------
    p_depolar_error_cnot : float, optional
        The probability of a depolarisation error after each CNOT gate.
        The default value is 0.
    single_qubit_gate_error_prob : float
        The probability of a depolarisation error occuring during a single
        qubit gate.
    meas_error_prob : float
        The probability of a bit flip error occuring during measurement, giving
        the opposite result from an ideal measurement.
    comm_qubit_depolar_rate : float, optional
        The depolarisation rate for communication qubits.
        The default value is 0.
    proc_qubit_depolar_rate : float
        The depolarisation rate for processing qubits.
        The default value is 0.
    single_qubit_gate_time : TYPE, optional
        DESCRIPTION. The default is 135 * 10**3.
    two_qubit_gate_time : TYPE, optional
        DESCRIPTION. The default is 600 * 10**3.
    measurement_time : TYPE, optional
        DESCRIPTION. The default is 600 * 10**4.
    alpha, beta : float, optional
        Settings for the artificial convenience gate INSTR_ARB_GEN, which
        initialises the state of a qubit as alpha |0> + beta |1>. The default
        is 1 for alpha and 0 for beta.
    num_positions : TYPE, optional
        DESCRIPTION. The default is 20.
    num_comm_qubits : TYPE, optional
        DESCRIPTION. The default is 2.

    Returns
    -------
    qprocessor : :class:`netsquid.components.qprocessor.QuantumProcessor

    Notes
    -----
    Uses universal gate sets from Ref. [1]_. Numerical noise is used.
    Gate durations are taken from averages for single and two-qubit gates on
    IonQ's Aria machine.

    References
    ----------
    .. [1] M. Nielsen and I. Chuang, Quantum Computation and Quantum
        Information, 10th ed. (Cambridge University Press, 2010).
    """

    def __init__(
        self,
        name="custom_noisy_qprocessor",
        p_depolar_error_cnot=0,
        single_qubit_gate_error_prob=0,
        meas_error_prob=0,
        comm_qubit_depolar_rate=0,
        proc_qubit_depolar_rate=0,
        single_qubit_gate_time=135 * 10**3,
        two_qubit_gate_time=600 * 10**3,
        measurement_time=600 * 10**4,
        alpha=1,
        beta=0,
        num_positions=20,
        num_comm_qubits=2,
    ):
        cnot_depolar_model = NDimDepolarNoiseModel(
            p_depolar_error_cnot, time_independent=True
        )
        comm_qubit_memory_depolar_model = DepolarNoiseModel(
            comm_qubit_depolar_rate, time_independent=False
        )
        processing_qubit_memory_depolar_model = DepolarNoiseModel(
            proc_qubit_depolar_rate, time_independent=False
        )
        meas_noise_model = MeasurementNoiseModel(meas_error_prob)
        single_qubit_gate_noise_model = DepolarNoiseModel(
            single_qubit_gate_error_prob, time_independent=True
        )
        # creating processor for all Nodes
        physical_instructions = (
            _phys_instructions_4_standard_lib_gates_and_convenience_ops(
                cnot_depolar_model,
                single_qubit_gate_noise_model,
                meas_noise_model,
                num_comm_qubits,
                single_qubit_gate_time,
                two_qubit_gate_time,
                measurement_time,
                alpha,
                beta,
            )
        )
        super().__init__(
            name,
            phys_instructions=physical_instructions,
            num_positions=num_positions,
            num_comm_qubits=num_comm_qubits,
            comm_qubit_mem_noise_model=comm_qubit_memory_depolar_model,
            processing_qubit_mem_noise_model=processing_qubit_memory_depolar_model,
        )

        self.add_composite_instruction(
            instr.INSTR_SWAP,
            [
                (instr.INSTR_CNOT, (0, 1)),
                (instr.INSTR_CNOT, (1, 0)),
                (instr.INSTR_CNOT, (0, 1)),
            ],
            topology=None,
        )
